from scipy import signal



class ButterLowpassFilter:
    """
    Butterworth lowpass filter.

    Args:
        cutoff (float): The cutoff frequency of the filter.
        fs (float, optional): The sampling frequency. Defaults to 250.
        order (int, optional): The order of the filter. Defaults to 2.

    Attributes:
        cutoff (float): The cutoff frequency of the filter.
        fs (float): The sampling frequency.
        order (int): The order of the filter.
        normal_cutoff (float): The normalized cutoff frequency.
        b (ndarray): The numerator coefficients of the filter.
        a (ndarray): The denominator coefficients of the filter.

    Methods:
        apply(data): Applies the lowpass filter to the input data.

    """

    """
    Initialize the ButterLowpassFilter.

    Args:
        cutoff (float): The cutoff frequency of the filter.
        fs (float, optional): The sampling frequency. Defaults to 250.
        order (int, optional): The order of the filter. Defaults to 2.
    """

    """
    Applies the lowpass filter to the input data.

    Args:
        data (ndarray): The input data to be filtered.

    Returns:
        ndarray: The filtered output data.
    """
    def __init__(self, cutoff, fs=250, order=2):
        self.cutoff = cutoff
        self.fs = fs
        self.order = order
        self.normal_cutoff = cutoff / (0.5 * fs)
        self.b, self.a = signal.butter(order, self.normal_cutoff, btype='low', analog=False)

    def apply(self, data):
        y = signal.filtfilt(self.b, self.a, data, axis=0)
        return y


class ButterHighpassFilter:
    """
    Initializes a Butterworth highpass filter.

    Parameters:
    - cutoff (float): The cutoff frequency of the filter.
    - fs (int, optional): The sampling frequency of the input data. Defaults to 250.
    - order (int, optional): The order of the filter. Defaults to 2.
    """

    """
    Applies the Butterworth highpass filter to the input data.

    Parameters:
    - data (ndarray): The input data to be filtered.

    Returns:
    - ndarray: The filtered data.
    """
    def __init__(self, cutoff, fs=250, order=2):
        self.cutoff = cutoff
        self.fs = fs
        self.order = order
        self.normal_cutoff = cutoff / (0.5 * fs)
        self.b, self.a = signal.butter(order, self.normal_cutoff, btype='high', analog=False)

    def apply(self, data):
        y = signal.filtfilt(self.b, self.a, data, axis=0)
        return y


class ButterBandpassFilter:
    """
    Initializes a Butterworth bandpass filter.

    Args:
        lowcut (float): The lower cutoff frequency of the filter.
        highcut (float): The upper cutoff frequency of the filter.
        fs (int, optional): The sampling frequency of the input data. Defaults to 250.
        order (int, optional): The order of the filter. Defaults to 2.
    """
    

    """
    Applies the Butterworth bandpass filter to the input data.

    Args:
        data (array-like): The input data to be filtered.

    Returns:
        array-like: The filtered output data.
    """

    def __init__(self, lowcut, highcut, fs=250, order=2):
        self.lowcut = lowcut
        self.highcut = highcut
        self.fs = fs
        self.order = order
        nyq = 0.5 * fs
        self.low = lowcut / nyq
        self.high = highcut / nyq
        self.b, self.a = signal.butter(order, [self.low, self.high], btype='band', analog=False)

    def apply(self, data):
        y = signal.filtfilt(self.b, self.a, data, axis=0)
        return y

class NotchFilter:
    """
    Initializes a notch filter.

    Args:
        cutoff (float): The cutoff frequency of the filter.
        fs (int, optional): The sampling frequency of the input data. Defaults to 250.
        Q (float, optional): The quality factor of the filter. Defaults to 30.

    """

    """
    Applies the notch filter to the input data.

    Args:
        data (array-like): The input data to be filtered.

    Returns:
        array-like: The filtered output data.
    """
    def __init__(self, cutoff, fs=250, Q=30):
        self.cutoff = cutoff
        self.fs = fs
        self.Q = Q
        self.b, self.a = signal.iirnotch(cutoff, Q, fs)

    def apply(self, data):
        y = signal.filtfilt(self.b, self.a, data, axis=0)
        return y


import pandas as pd
import os


def load_csv(file_path):
    # df : Elapsed Time,Isquio,Cuadriceps,GLMedio,AductorLargo,Angle
    # Load csv
    df = pd.read_csv(file_path)
    # Set dtypes
    df = df.astype(float)
    return df


def load_data(path):
    files = os.listdir(path)
    files = [f for f in files if f.endswith(".csv")]
    files.sort()

    data = [load_csv(os.path.join(path, f)) for f in files]
    tiempo_total = sum([len(df)/(250*3600) for df in data])
    print(f"Total time: {tiempo_total:.2f} hours")
    return data





import os; os.system('clear')
import argparse
import logging
from brainflow.board_shim import BoardShim, BrainFlowInputParams, BoardIds
import time
from paho.mqtt import client as mqtt_client
import ciervo.parameters as p
import serial
import sys
import glob
from pprint import pprint
from ciervo.aux_tools import Buffer
import numpy as np
from time import sleep

# Import SciPy signal processing functions
from scipy.signal import iirnotch, butter, sosfilt, tf2sos

def create_channel_setting_command(channel, power_down, gain_set, input_type_set, bias_set, srb2_set, srb1_set, daisy_module=False):
    """
    Generate the channel setting command string.
    """
    if not (1 <= channel <= 16):
        raise ValueError("Channel must be between 1 and 16.")
    
    # Determine channel letter/number based on board or daisy module
    if channel <= 8:
        channel_str = str(channel)
    else:
        channel_str = chr(ord('Q') + (channel - 9))  # 'Q' to 'I' for daisy module channels

    # Construct the command string
    command = f"x{channel_str}{power_down}{gain_set}{input_type_set}{bias_set}{srb2_set}{srb1_set}x"

    return command

def find_openbci():
    """ Lists serial port names """
    if sys.platform.startswith('win'):
        ports = ['COM%s' % (i + 1) for i in range(256)]
    elif sys.platform.startswith('linux') or sys.platform.startswith('cygwin'):
        # Exclude current terminal "/dev/tty"
        ports = glob.glob('/dev/tty[A-Za-z]*')
    elif sys.platform.startswith('darwin'):
        ports = glob.glob('/dev/tty.*')
    else:
        raise EnvironmentError('Unsupported platform')

    result = []
    for port in ports:
        try:
            s = serial.Serial(port)
            s.close()
            result.append(port)
        except (OSError, serial.SerialException):
            pass

    open_bci_ports = [port for port in result if 'usb' in port.lower()]
    #print(open_bci_ports)
    #open_bci_ports = ['/dev/ttyAMA1']

    assert len(open_bci_ports) > 0, 'No se encontraron dispositivos OpenBCI'
    assert len(open_bci_ports) == 1, 'Se encontraron varios dispositivos OpenBCI'

    print(f"OpenBCI en puerto {open_bci_ports[0]}")

    return open_bci_ports[0]

class Publish:
    def __init__(self, board_shim):
        self.board_id = board_shim.get_board_id()
        self.board_shim = board_shim
        self.exg_channels = BoardShim.get_exg_channels(self.board_id)
        self.sampling_rate = BoardShim.get_sampling_rate(self.board_id)
        self.update_speed_ms = 50
        self.window_size = 5
        self.num_points = self.window_size * self.sampling_rate

        self.marker = -1
        self.buffer = Buffer(self.window_size, roll=True)


        # Initialize filters using SciPy
        self.mains = 50  # Notch filter frequency (Hz)
        self.band_low = 20  # Lower cutoff frequency for band-pass filter (Hz)
        self.band_high = 120  # Upper cutoff frequency for band-pass filter (Hz)

        # Design notch filter
        b_notch, a_notch = iirnotch(w0=self.mains / (self.sampling_rate / 2), Q=30)
        self.sos_notch = tf2sos(b_notch, a_notch)

        # Design band-pass filter
        self.sos_bp = butter(
            N=4,
            Wn=[self.band_low / (self.sampling_rate / 2), self.band_high / (self.sampling_rate / 2)],
            btype='bandpass',
            output='sos'
        )

        # Initialize filter states for each EEG channel
        self.num_eeg_channels = len(self.exg_channels)
        self.zi_bp = []
        self.zi_notch = []
        for _ in range(self.num_eeg_channels):
            self.zi_bp.append(np.zeros((self.sos_bp.shape[0], 2)))
            self.zi_notch.append(np.zeros((self.sos_notch.shape[0], 2)))

        # MQTT
        self.broker = p.BROKER_HOST
        self.port = p.BROKER_PORT
        self.topic = 'data'
        self.client = mqtt_client.Client(mqtt_client.CallbackAPIVersion.VERSION2, client_id='openbci')
        self.client.on_connect = on_connect
        self.client.on_message = self.on_message

        self.client.connect(self.broker, self.port)
        self.client.subscribe('marker')  # Subscribe to marker topic

        self.client.loop_start()

        self.update()
        self.client.loop_stop()

    def on_message(self, client, userdata, msg):
        self.marker = int(msg.payload.decode('utf-8'))

    def update(self):
        target_set = set(list(range(256)))
        start_time = None
        while True:
            time.sleep(8 / self.sampling_rate)  # Wait for at least 2 samples
            data = self.board_shim.get_board_data(self.num_points)  # np.float64 default
            if data.shape[1] == 0:
                continue

            if start_time is None:
                start_time = data[p.TIME_CHANNEL, 0]
            data[p.TIME_CHANNEL, :] -= start_time

            if self.marker != -1:
                data[p.MARKER_CHANNEL, :] = self.marker
            data = data[p.ALL_CHANNELS, :]

            # Apply filters to EEG channels
            for idx, channel in enumerate(range(8)):
                # Get data for this channel
                channel_data = data[channel, :] - 2250

                # Apply the band-pass filter with state
                filtered_bp, self.zi_bp[idx] = sosfilt(
                    self.sos_bp, channel_data, zi=self.zi_bp[idx]
                )

                # Apply the notch filter with state
                filtered_data, self.zi_notch[idx] = sosfilt(
                    self.sos_notch, filtered_bp, zi=self.zi_notch[idx]
                )

                # Assign filtered data back to data array
                data[channel, :] = filtered_data

            # Convert data to desired precision and publish
            data = data.astype(p.PRECISION)
            #print(data.shape)
            data_bytes = data.tobytes()

            
            obtain_set = set(data[12, :])

            if not obtain_set.issubset(target_set):
                print("Reiniciando")
                self.board_shim.release_session()
                self.board_shim.prepare_session()
                self.board_shim.start_stream()



            
            
             

            self.client.publish(self.topic, data_bytes, qos=0)

def on_connect(client, userdata, flags, rc, properties):
    print(f"Connected with result code {rc}")

def main():
    BoardShim.enable_dev_board_logger()
    logging.basicConfig(level=logging.DEBUG)

    parser = argparse.ArgumentParser()
    # Use docs to check which parameters are required for specific board, e.g., for Cyton - set serial port
    parser.add_argument('--timeout', type=int, help='timeout for device discovery or connection', required=False,
                        default=0)
    parser.add_argument('--ip-port', type=int, help='ip port', required=False, default=0)
    parser.add_argument('--ip-protocol', type=int, help='ip protocol, check IpProtocolType enum', required=False,
                        default=0)
    parser.add_argument('--ip-address', type=str, help='ip address', required=False, default='')
    parser.add_argument('--mac-address', type=str, help='mac address', required=False, default='')
    parser.add_argument('--other-info', type=str, help='other info', required=False, default='')
    parser.add_argument('--streamer-params', type=str, help='streamer params', required=False, default='streaming_board://225.1.1.1:6677')
    parser.add_argument('--serial-number', type=str, help='serial number', required=False, default='')
    parser.add_argument('--board-id', type=int, help='board id, check docs to get a list of supported boards',
                        required=False, default=BoardIds.CYTON_BOARD)
    parser.add_argument('--file', type=str, help='file', required=False, default='')
    parser.add_argument('--master-board', type=int, help='master board id for streaming and playback boards',
                        required=False, default=BoardIds.CYTON_BOARD)
    args = parser.parse_args()

    params = BrainFlowInputParams()
    params.ip_port = args.ip_port
    params.serial_port = find_openbci()
    params.mac_address = args.mac_address
    params.other_info = args.other_info
    params.serial_number = args.serial_number
    params.ip_address = args.ip_address
    params.ip_protocol = args.ip_protocol
    params.timeout = args.timeout
    params.file = args.file
    params.master_board = args.master_board
    #params.ip_address_aux = "225.1.1.2"
    #params.ip_port_aux = 6678

    board_shim = BoardShim(args.board_id, params)

    try:
        board_shim.prepare_session()
        # Configure OpenBCI
        # x (CHANNEL, POWER_DOWN, GAIN_SET, INPUT_TYPE_SET, BIAS_SET, SRB2_SET, SRB1_SET) X

        for i in range(1, 9):
            ch = create_channel_setting_command(channel=i,
                                                power_down= 0 if i >4 else 1,
                                                gain_set=6,
                                                input_type_set=0,
                                                bias_set=1,
                                                srb2_set=1,
                                                srb1_set=0)
            

            board_shim.config_board(ch)
            sleep(0.1)

        #board_shim.add_streamer(args.streamer_params)
        board_shim.start_stream()
        Publish(board_shim)
    except BaseException:
        logging.warning('Exception', exc_info=True)
    finally:
        logging.info('End')
        if board_shim.is_prepared():
            logging.info('Releasing session')
            board_shim.release_session()

if __name__ == '__main__':
    board_id = BoardIds.CYTON_BOARD.value
    pprint(BoardShim.get_board_descr(board_id))

    main()



from typing import Union
import pandas as pd
import numpy as np





def train_test_split(data: Union[list, pd.DataFrame], columna: list,  window_size=125, test_size=0.2, overlap=0,  random_state=None):
    """
    Splits the data into training and testing sets based on the specified window size and test size.
    Parameters:
    - data (Union[list, pd.DataFrame]): The data to be split. It can be either a list of dataframes or a single dataframe.
    - window_size (int): The size of each window for splitting the data. Default is 125.
    - test_size (float): The proportion of data to be allocated for testing. Default is 0.2.
    - random_state (int): The seed value for random number generation. Default is None.
    Returns:
    - train (numpy.ndarray): The training data. (N, window_size, columns)
    - test (numpy.ndarray): The testing data. (N, window_size, columns)
    """

    if random_state:
        np.random.seed(random_state)

    
    if isinstance(data, pd.DataFrame):
        data = [data]

    train_data = []
    train_label = []

    test_data = []
    test_label = []


    columna = [c.strip() for c in columna]

    for df_idx, df in enumerate(data):

        # check if columna is in df.columns
        column_is_present = True
        for c in columna:
            if c not in df.columns:
                print(f"El archivo id: {df_idx} no cuenta con la columna {c}. Archivo ignorado")
                column_is_present = False

        # Ignora archivo
        if not column_is_present:
            continue

        values = df[columna]
        label = df['labels']

        if overlap == 0:
            # non overlapping
            # number of windows
            n_windows = int(len(df)/window_size)

            # random indices
            indices = np.random.choice(range(n_windows), int(n_windows*test_size), replace=False)

            # split data
            for i in range(n_windows):
                if i in indices:
                    test_data.append(values.iloc[i*window_size:(i+1)*window_size])
                    test_label.append(label.iloc[(i+1)*window_size])
                else:
                    train_data.append(values.iloc[i*window_size:(i+1)*window_size])
                    train_label.append(label.iloc[(i+1)*window_size])

        else:
            # overlaping
            split_df = np.array_split(df, 20)

            # random indices
            test_indices = np.random.choice(range(len(split_df)), int(len(split_df)*test_size), replace=False)


            for s_idx, s_df in enumerate(split_df):
                n_windows = int((len(s_df)- window_size) // overlap)
                
                for i in range(n_windows):
                    if s_idx in test_indices:
                        test_data.append(values.iloc[i*overlap:(i+1)*overlap])
                        test_label.append(label.iloc[(i+1)*overlap])
                    else:
                        train_data.append(values.iloc[i*overlap:(i+1)*overlap])
                        train_label.append(label.iloc[(i+1)*overlap])


            

    train_data = np.array(train_data)
    train_label = np.array(train_label)

    test_data = np.array(test_data)
    test_label = np.array(test_label)

    
    return train_data, train_label, test_data, test_label



if __name__ == '__main__':
    import os; os.system('clear')

    from ciervo.io.load import example_marcha_larga

    data = example_marcha_larga()

    train,test = train_test_split(data, test_size=0.1, random_state=42)
    print(train.shape, test.shape)


    


import os; os.system('clear')
import numpy as np
from scipy import signal
import scipy
from ciervo.models import train_test_split
import pandas as pd
import ciervo.parameters as p
from ciervo.models import features_v1
from sklearn.ensemble import RandomForestClassifier
from sklearn.svm import SVC
from sklearn.linear_model import Perceptron
from sklearn.neural_network import MLPClassifier
from sklearn.metrics import accuracy_score
from sklearn.model_selection import GridSearchCV
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler
from matplotlib import pyplot as plt
import joblib



#otro6, otro5, viernes, last_day2, last_day
def remove_nan_values(features, labels):
    """Remove values with NaN values"""
    isnan = np.isnan(features).any(axis=1)

    #features[isnan] = 0


    return features, labels

    


if __name__ == "__main__":
    # Load data
    files = ['caminata_normal']

    data = []
    for file in files:
        temp = np.load(f'recordings/{file}.npy')
        print(temp.shape)
        data.append(temp)

    data = np.hstack(data)



    n_points = data.shape[1]
    print(f"Recording time: {n_points/250} s")

    df = pd.DataFrame(data.T, columns=p.CH_NAMES)
    df['labels'] = df['MARKERS']

    print(set(df['labels'].tolist()))

    df = df[df['labels'] != 0]
    df['labels'] = df['labels'].apply(lambda x: 1 if x == 2 else 0)

    fig = plt.figure()

    plt.plot(df['labels'].to_numpy(), label='orig')

    # Smooth filter
    kernel_size = 200 
    #df['labels'] = np.convolve(df['labels'].to_numpy(), np.array([1]*kernel_size) / kernel_size, mode='same')
    #df['labels'] = np.convolve(df['labels'].to_numpy(), np.array([1]*kernel_size) / kernel_size, mode='same')
    #df['labels'] = np.convolve(df['labels'].to_numpy(), np.array([1]*kernel_size) / kernel_size, mode='same')

    plt.plot(df['labels'].to_numpy(), label='filter')

    #df['labels'] = df['labels'].apply(lambda x: 1 if x > 0.5 else 0)

    plt.plot(df['labels'].to_numpy(), label='final')
    plt.legend()

    plt.show()

    train_data, train_label, test_data, test_label = train_test_split([df], 
                                         columna=['C1', 'C2', 'C3', 'C4'],
                                         window_size=25,
                                         overlap=0, 
                                         test_size=0.2,
                                         random_state=42)
    
    print(train_data.shape, train_label.shape, test_data.shape, test_label.shape)

    # Feature extraction
    train_features = [features_v1(x.T)[0] for x in train_data]
    train_features = np.array(train_features)

    test_features = [features_v1(x.T)[0] for x in test_data]
    test_features = np.array(test_features)

    # Remove trials with NaN values
    train_features, train_label = remove_nan_values(train_features, train_label)
    test_features, test_label = remove_nan_values(test_features, test_label)


    print(train_features.shape, test_features.shape)
    print(train_label.shape, test_label.shape)

    # Classifier and parameter grid setup
    classifiers = {
        'RandomForest': RandomForestClassifier()
       # 'SVC': make_pipeline(StandardScaler(), SVC())
    }

    param_grids = {
        'RandomForest': {
            'n_estimators': [10, 20, 30 ,40 ,50, 80, 100],
            'max_depth': [None, 10, 20],
            'min_samples_split': [2, 5],
        }
        #'SVC': {
        #    'svc__C': [0.1, 1, 10],
        #    'svc__kernel': ['linear', 'rbf'],
        #}
    }

    best_accuracy = 0.0
    best_clf = None
    best_name = ""

    for clf_name in classifiers:
        clf = classifiers[clf_name]
        param_grid = param_grids[clf_name]
        
        grid_search = GridSearchCV(clf, param_grid, cv=3, scoring='accuracy')
        grid_search.fit(train_features, train_label)

        print(f"{clf_name} best parameters: {grid_search.best_params_}")
        accuracy = accuracy_score(test_label, grid_search.predict(test_features))
        print(f"{clf_name} test accuracy: {accuracy * 100:.2f}%")

        if accuracy > best_accuracy:
            best_accuracy = accuracy
            best_clf = grid_search.best_estimator_
            best_name = clf_name

    print(f"Best classifier: {best_name} with accuracy: {best_accuracy * 100:.2f}%")
    
    # Save the best model
    joblib.dump(best_clf, 'model.pkl')

    # Load and test the saved best model
    loaded_clf = joblib.load('model.pkl')
    y_pred = loaded_clf.predict(test_features)
    print(f"Loaded model accuracy: {accuracy_score(test_label, y_pred) * 100:.2f}%")


import os; os.system('clear')
import pandas as pd
import argparse
import numpy as np
from matplotlib import pyplot as plt
from ciervo.plots import emg_plot
from ciervo.io import example_marcha_larga, example_marcha



def main():
    parser = argparse.ArgumentParser(description='Fix BioRadio CSV file')
    parser.add_argument('--input', '-i', type=str, help='Input CSV file')
    # bool
    parser.add_argument('--bad_decimal', action='store_true', help='Se usa si el csv tiene decimales con coma')
    args = parser.parse_args()


    if args.bad_decimal:
        with open(args.input, 'r') as f:
            lines = f.readlines()
            columns = lines[0].split(',')
            # remove from columns ,'BioRadio Event', '\n'
            print(columns)
            columns = columns[:-2]
            data = []
            bad_count = 0
            for line in lines[1:]:
                split_line = line.split(',')
                #print(split_line)
                #print(len(split_line))
                row = [split_line[0]]
                for i in range(1, len(split_line) -2, 2):
                    try:
                        v = float('.'.join(split_line[i:i+2]))
                        row.append(float('.'.join(split_line[i:i+2])))
                    except:
                        bad_count += 1
                        pass

                data.append(row)
            df = pd.DataFrame(data, columns=columns)
            print(df.head())
            # Percentaje of bad values
            print(f'Bad values: {bad_count/len(data)*100:.2f}%')


    else:
        df = pd.read_csv(args.input, sep=',', decimal='.')
        #df = df.drop(columns=['BioRadio Event', 'Unnamed: 13'])



    # rename columns
    df = df.rename(columns={
    'Grip Strength Gionomet': 'Angle',
    'Flexlsq': 'EMG_Isquio',
    'ExtCuad': 'EMG_Cuadriceps',
    'Aductor': 'EMG_AductorLargo',
    'AbdGluM': 'EMG_GLMedio'
    })
    
    # convertir Angulo a grados
    df['Angle'] = 0.5107* df['Angle'] -216.03  # Convert to degrees

    # Convert formato de tiempo a segundos
    df['Elapsed Time'] = np.linspace(0, len(df['Elapsed Time'])/250, len(df['Elapsed Time']))
    print("Total duration in minutes: ", len(df['Elapsed Time'])/(60*250))

    # Todo a float
    df = df.astype(float)

    # Asign nan to values above threshold
    df = df.mask(df > 1000)

    # interpolate nan values
    df = df.interpolate()



if __name__ == '__main__':
    main()
    

